"""PolicyGlass."""
from .action import Action, EffectiveAction
from .condition import Condition, ConditionCollection, ConditionKey, ConditionOperator, ConditionValue
from .policy import Policy
from .policy_shard import PolicyShard, dedupe_policy_shards, policy_shards_effect
from .principal import EffectivePrincipal, Principal, PrincipalCollection, PrincipalType, PrincipalValue
from .resource import EffectiveResource, Resource
from .statement import Statement

__all__ = [
    "Policy",
    "Statement",
    "Principal",
    "Action",
    "Resource",
    "PrincipalCollection",
    "Principal",
    "PrincipalType",
    "PrincipalValue",
    "Condition",
    "ConditionKey",
    "ConditionOperator",
    "ConditionValue",
    "ConditionCollection",
    "EffectiveAction",
    "EffectiveResource",
    "EffectivePrincipal",
    "PolicyShard",
    "policy_shards_effect",
    "dedupe_policy_shards",
]



"""Resource class."""
from fnmatch import fnmatch
from typing import List

from .effective_arp import EffectiveARP


class Resource(str):
    """A resource ARN may be case sensitive or case insensitive depending on the resource type."""

    def __lt__(self, other: object) -> bool:
        """Whether this object contains (but is not equal to) another object.

        Parameters:
            other: The object to determine if our object contains.

        Raises:
            ValueError: If the other object is not of the same type as this object.
        """
        if not isinstance(other, self.__class__):
            raise ValueError(f"Cannot compare {self.__class__.__name__} and {other.__class__.__name__}")
        if self == other:
            return False
        return fnmatch(self, other)

    def issubset(self, other: object) -> bool:
        """Whether this object contains all the elements of another object (i.e. is a subset of the other object).

        Parameters:
            other: The object to determine if our object contains.

        Raises:
            ValueError: If the other object is not of the same type as this object.
        """
        if not isinstance(other, self.__class__):
            raise ValueError(f"Cannot compare {self.__class__.__name__} and {other.__class__.__name__}")
        for self_element, other_element in zip(self.arn_elements, other.arn_elements):
            if not fnmatch(self_element, other_element):
                return False
        return True

    @property
    def arn_elements(self) -> List[str]:
        """Return a list of arn elements, replacing blanks with ``*``."""
        return [element or "*" for element in self.split(":")]

    def __contains__(self, other: object) -> bool:
        """Not Implemented.

        Parameters:
            other: The object to see if this object contains.

        Raises:
            NotImplementedError: This method is not implemented.
        """
        raise NotImplementedError()

    def __repr__(self) -> str:
        """Return an instantiable representation of this object."""
        return f"{self.__class__.__name__}('{self}')"


class EffectiveResource(EffectiveARP[Resource]):
    """EffectiveResources are the representation of the difference between an Resource and its exclusion.

    The allowed Resource is the difference (subtraction) of the excluded Resources
    from the included Resource.
    """

    _arp_type = Resource



"""Statement class."""

from typing import Dict, FrozenSet, List, Optional, Tuple, TypeVar, Union

from pydantic import BaseModel, validator

from .action import Action, EffectiveAction
from .condition import Condition, ConditionCollection, ConditionKey, ConditionOperator, ConditionValue
from .policy_shard import PolicyShard
from .principal import EffectivePrincipal, Principal, PrincipalCollection, PrincipalType, PrincipalValue
from .resource import EffectiveResource, Resource
from .utils import to_pascal

T = TypeVar("T")


class Effect(str):
    """Allow or Deny."""

    ...


class Statement(BaseModel):
    """A Policy Statement."""

    effect: Effect
    action: Optional[List[Action]]
    not_action: Optional[List[Action]]
    resource: Optional[List[Resource]]
    not_resource: Optional[List[Resource]]
    principal: Optional[PrincipalCollection]
    not_principal: Optional[PrincipalCollection]
    condition: Optional[ConditionCollection]

    class Config:
        """Configure the Pydantic BaseModel."""

        alias_generator = to_pascal

    def policy_json(self) -> str:
        return self.json(by_alias=True, exclude_none=True)

    @property
    def policy_shards(self) -> List[PolicyShard]:
        conditions: FrozenSet[Condition] = frozenset({})
        not_principals: FrozenSet[Principal] = frozenset({})
        if self.condition:
            conditions = frozenset(self.condition.conditions)
        if self.not_principal:
            not_principals = frozenset(self.not_principal.principals)
        result = []
        arps = []
        for action in self.action or [Action("*")]:
            if self.principal:
                principals = self.principal.principals
            else:
                principals = [Principal(PrincipalType("AWS"), PrincipalValue("*"))]
            for principal in principals:
                for resource in self.resource or [Resource("*")]:
                    arp: Tuple[Action, Resource, Principal] = (action, resource, principal)
                    arps.append(arp)

        for action, resource, principal in arps:
            result.append(
                PolicyShard(
                    effect=self.effect,
                    effective_action=EffectiveAction(
                        action,
                        exclusions=frozenset(self.not_action or {}),
                    ),
                    effective_resource=EffectiveResource(
                        resource,
                        exclusions=frozenset(self.not_resource or {}),
                    ),
                    effective_principal=EffectivePrincipal(principal, exclusions=not_principals),
                    conditions=conditions,
                )
            )
        return result

    @validator("action", "not_action", pre=True)
    def ensure_action_list(cls, v: T) -> List[Action]:
        if isinstance(v, list):
            return [Action(action) for action in v]
        return [Action(v)]

    @validator("resource", "not_resource", pre=True)
    def ensure_resource_list(cls, v: T) -> List[Resource]:
        if isinstance(v, list):
            return [Resource(resource) for resource in v]
        return [Resource(v)]

    @validator("condition", pre=True)
    def ensure_condition_value_list(
        cls, v: Dict[ConditionKey, Dict[ConditionOperator, Union[ConditionValue, List[ConditionValue]]]]
    ) -> ConditionCollection:
        output: Dict = {}
        for operator, key_and_values in v.items():
            output[operator] = {}
            for key, values in key_and_values.items():
                if isinstance(values, list):
                    output[ConditionOperator(operator)][ConditionKey(key)] = values
                else:
                    output[ConditionOperator(operator)][ConditionKey(key)] = [values]
        return ConditionCollection(output)

    @validator("principal", "not_principal", pre=True)
    def ensure_principal_dict(
        cls, v: Union[PrincipalValue, Dict[PrincipalType, Union[PrincipalValue, List[PrincipalValue]]]]
    ) -> PrincipalCollection:
        if not isinstance(v, dict):
            return PrincipalCollection({PrincipalType("AWS"): [PrincipalValue(v)]})
        output = dict()
        for principal_type, principals in v.items():
            if isinstance(principals, list):
                output[principal_type] = principals
            else:
                output[principal_type] = [principals]
        return PrincipalCollection(output)



"""Utilities for PolicyGlass."""


def to_pascal(string: str) -> str:
    """Convert a snake_case string into a PascalCase string.

    Parameters:
        string: The string to convert to PascalCase.
    """
    return "".join(word.capitalize() for word in string.split("_"))



#!/usr/bin/env python
"""Setup CloudWanderer package."""
import re
from os import path

from setuptools import find_packages, setup

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.rst"), encoding="utf-8") as f:
    long_description = re.sub(r"..\s*doctest\s*::", ".. code-block ::", f.read())

long_description = re.sub(r":class:`~[^`]+\.([^`]+)`", "\1", long_description)

setup(
    version="0.2.1",
    python_requires=">=3.6.0",
    name="policyglass",
    packages=find_packages(include=["policyglass", "policyglass.*"]),
    description="Understand the effective permissions of your policies",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    author="Sam Martin",
    author_email="samjackmartin+policyglass@gmail.com",
    url="https://github.com/CloudWanderer-io/PolicyGlass",
    install_requires=["pydantic", 'typing_extensions; python_version < "3.8.0"'],
    package_data={
        "": ["py.typed"],
    },
)
